"""
Abstract base class for ring elements.
"""

##################################################################
# Generic ring element, so all this functionality must be defined
# by any ring element.  Derived class must call __init__
##################################################################

import operator

import sage.structure.element_py

import sage.ext.coerce

# This classes uses element.pxd.  To add data members, you
# must change that file. 

cdef class Element:
    """
    Generic element base class, so all this functionality must
    be defined by any ring element.
    """

    def __init__(self, parent):
        self.__parent = parent

    def __reduce__(self):
        return sage.rings.element_py.init, (self.parent(),)

    def __repr__(self):
        return "Generic element of a ring"

    def __hash__(self):
        return hash(str(self))

    def base_ring(self):
        return self.parent().base_ring()

    def category(self):
        from sage.categories.categories import Elements
        return Elements(self.parent())

    def is_zero(self):
        return self == self.__parent(0)

    def is_nonzero(self):
        return not self.is_zero()

    def parent(self, x=None):
        if x==None:
            return self.__parent
        else:
            return self.__parent(x)

    def name(self):
        return self.__parent.name()

    def names(self):
        return self.__parent.names()

    def __xor__(self, right):
        raise RuntimeError, "Use ** for exponentiation, not '^', which means xor\n"+\
              "in Python, and has the wrong precedence."

    def __add__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.add)
        return self._add(right)
    def _add(self, right):
        raise NotImplementedError

    def __sub__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.sub)
        return self._sub(right)
    def _sub(self, right):
        return self + (-right)

    def __mul__(self, right):
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.mul)
        return self._mul(right)
    
    def _mul(self, right):
        raise NotImplementedError
    

    def __div__(self, right):
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.div)
        return self._div(right)
    def _div(self, right):
        raise NotImplementedError
    
    def __neg__(self):
        return sage.ext.coerce.bin_op(-1, self, operator.mul)

    def __pos__(self):
        return self



    
cdef class ModuleElement(Element):
    """
    Generic element of a module.
    """
    pass

cdef class MonoidElement(Element):
    """
    Generic element of a monoid.
    """

cdef class RingElement(Element):
    def is_unit(self):
        if self == 1 or self == -1:
            return True
        raise NotImplementedError
    
    def additive_order(self):
        return NotImplementedError

    def multiplicative_order(self):
        return NotImplementedError

    def __pow__(self, n, dummy):
        cdef int i
        a = self
        power = self.parent()(1)
        if n < 0:
            n = -n
            a = ~self
        elif n == 0:
            return power
        i = 0
        apow2 = a
        while (n>>i) > 0:
            if (n>>i) & 1:
                power = power * apow2
            if n == 0: break   # to not waste time doing an extra multiplication/increment
            apow2 = apow2 * apow2
            i = i+1
        return power

    def _div(self, right):
        return self.parent().fraction_field()(self, right)



cdef class CommutativeRingElement(RingElement):
    pass

cdef class IntegralDomainElement(CommutativeRingElement):
    pass

cdef class DedekindDomainElement(IntegralDomainElement):
        
    pass
        
cdef class PrincipalIdealDomainElement(DedekindDomainElement):
    def gcd(self, other):
        raise NotImplementedError
##         R = self.parent()
##         I = R.ideal(self)
##         J = R.ideal(other)
##         K = I + J
##         return K.gen()
        

cdef class EuclideanDomainElement(PrincipalIdealDomainElement):

    def degree(self):
        raise NotImplementedError

    def gcd(self, other):
        """
        Return the greatest common divisor of self and other.

        Algorithm 3.2.1 in Cohen, GTM 138.
        """
        A = self
        B = other
        while B.is_nonzero():
            Q, R = A.quo_rem(B)
            A = B
            B = R
        return A

    def leading_coefficient(self):
        raise NotImplementedError

    def quo_rem(self, other):
        raise NotImplementedError

    def __floordiv__(self,right):
        """
        Quotient of division of self by other.  This is denoted //.
        """
        Q, _ = self.quo_rem(right)
        return Q
        
    def __mod__(self, other):
        """
        Remainder of division of self by other.
        EXAMPLES:
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: x % (x+1)
            -1
            sage: (x**3 + x - 1) % (x**2 - 1)
            2*x - 1
        """
        _, R = self.quo_rem(other)
        return R
        
    

cdef class FieldElement(RingElement):

    def is_unit(self):
        return not self.is_zero()
    
    def gcd(self, other):
        """
        Return the greatest common divisor of self and other.
        """
        if not isinstance(other, FieldElement) or other.parent() != self:
            other = self.parent()(other)
        if self.is_zero() and other.is_zero():
            return self
        else:
            return self.parent()(1)

    def quo_rem(self, right):
        if not isinstance(right, FieldElement) or right.parent() != self:
            right = self.parent()(right)
        return self/right, 0

    
cdef class AlgebraElement(RingElement):
    pass

cdef class CommutativeAlgebraElement(AlgebraElement):
    pass    

cdef class InfinityElement(RingElement):
    pass
